/*-
 * Copyright (c) 2000 Munehiro Matsuda
 * Copyright (c) 2000 Mitsuru IWASAKI <iwasaki@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$FreeBSD$
 */

#include <sys/param.h>
#include <sys/ioctl.h>

#include <dev/acpica/acpiio.h>

#include <err.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define ACPIDEV	"/dev/acpi"

#define BATT_STAT_DISCHARG	0x01
#define BATT_STAT_CHARGING	0x02
#define BATT_STAT_CRITICAL	0x04
#define BATT_STAT_MAX		   7

#define BATT_AC_ONLINE		0x08

#define BATT_INFO_MAX		BATT_AC_ONLINE

struct cmbat_info {
	struct acpi_bif bif;
	struct acpi_bst bst;
	int	cap;
	int	min;
};

void
usage(void)
{
	fprintf(stderr, "usage: acpibatt [-vd?]\n");
	exit(1);
}

static int
acpi_acad_status(void)
{
	int	fd;
	int	status;

	fd  = open(ACPIDEV, O_RDONLY);
	if (fd == -1) {
		err(1, "%s", ACPIDEV);
	}
	if (ioctl(fd, ACPIIO_ACAD_GET_STATUS, &status) == -1) {
		err(1, "Failed to get AC-Line info");
	}
	close(fd);

	return(status);
}

static void
acpi_cmbat_get_units(int *unitsp)
{
	int	fd;

	fd  = open(ACPIDEV, O_RDONLY);
	if (fd == -1) {
		err(1, "%s", ACPIDEV);
	}
	if (ioctl(fd, ACPIIO_CMBAT_GET_UNITS, unitsp) == -1) {
		err(1, "Failed to get number of batteries");
	}
	close(fd);
}

static void
acpi_cmbat_get_info(int unit, struct acpi_bif *bifp, struct acpi_bst *bstp)
{
	int	 fd;
	union	 acpi_cmbat_ioctl_arg arg;

	fd  = open(ACPIDEV, O_RDONLY);
	if (fd == -1) {
		err(1, "%s", ACPIDEV);
	}
	arg.unit = unit;
	if (ioctl(fd, ACPIIO_CMBAT_GET_BIF, &arg) == -1) {
		err(1, "Failed to get ACPI _BIF info");
	}
	bcopy(&arg.bif, bifp, sizeof(struct acpi_bif));

	arg.unit = unit;
	if (ioctl(fd, ACPIIO_CMBAT_GET_BST, &arg) == -1) {
		err(1, "Failed to get ACPI _BST info");
	}
	bcopy(&arg.bst, bstp, sizeof(struct acpi_bst));

	close(fd);
}

int
main(int argc, char *argv[])
{
	int i, c;
	int stat;
	int debug;
	int units;
	int verbose;
	int total_cap, total_min;
	int valid_rate, valid_units;
	struct cmbat_info *bat;
	char *status[BATT_INFO_MAX] = {
		"normal",
		"dis-charging",
		"charging",
		NULL,
		"critical",
		"critical and dis-charging",
		"critical and charging",
		"not present"
	};
	char *power_units[]= { "mWh", "mAh" };

	verbose = debug = 0;
	while ((c = getopt(argc, argv, "vd?")) != -1) {
		switch (c) {
		case 'd':
			debug++;
			break;
		case 'v':
			verbose++;
			break;
		case '?':
		default:
			usage();
		}
		argc -= optind;
		argv += optind;
	}

	if (acpi_acad_status()) {
		stat |= BATT_AC_ONLINE;
	}
	acpi_cmbat_get_units(&units);

	bat = (struct cmbat_info *)malloc(sizeof(struct cmbat_info) * units);
	if (bat == NULL) {
		err(1, "malloc failed");
	}

	stat = valid_rate = valid_units = 0;
	for (i = 0; i < units; i++) {
		acpi_cmbat_get_info(i, &bat[i].bif, &bat[i].bst);

		/* If battey not installed, state becomes 0xffff */
		if (bat[i].bst.state >= BATT_STAT_MAX) {
			bat[i].bst.state = BATT_STAT_MAX;
			continue;
		}
		valid_units++;

		bat[i].cap = 100* bat[i].bst.cap / bat[i].bif.lfcap;

		if (bat[i].bst.state & BATT_STAT_CHARGING) {
			stat |= BATT_STAT_CHARGING;
		} else
		if (bat[i].bst.rate > 0) {
			/*
			 * XXX Hack to calculate total battery time.
			 * Systems with two or more battries, they get used
			 * one by one, thus bst.rate is set only to the one
			 * in use. For remaining batteries bst.rate = 0, which
			 * makes it impossible to calculate remaining time.
			 *
			 * There for we find the largest bst.rate that is
			 * valid (the one in dis-charging state), and use its
			 * value to calcutate remaining batteries' time.
			 */
			if ((bat[i].bst.state & BATT_STAT_DISCHARG) &&
			    (valid_rate < bat[i].bst.rate)) {
				valid_rate = bat[i].bst.rate;
			}
		}
	}

	/* Check AC-Line status */
	if (acpi_acad_status()) {
		stat |= BATT_AC_ONLINE;
	}

	/* Calculate total battery capacity and time */
	total_min = total_cap = 0;
	for (i = 0; i < units; i++) {
		if (bat[i].bst.state >= BATT_STAT_MAX)
			break;

		if (bat[i].bst.rate > 0) {
			/* Do the proper way */
			bat[i].min = 60 * bat[i].bst.cap / bat[i].bst.rate;
#if 0
			bat[i].min /= units;
#endif
		} else {
			if (valid_rate != 0) {
				/* We've found valid bst.rate, let's use it */
				bat[i].min = 60 * bat[i].bst.cap / valid_rate;
				bat[i].bst.rate = valid_rate;
			} else {
				/* Couldn't find valid bst.rate */
				bat[i].min = 0;
			}
		}
		total_min += bat[i].min;
		total_cap += bat[i].cap;
	}

	/* Print Battery related informations */
	printf("ACPI Battey info:\n");
#if 0
	printf("ACPI version:\n");
	printf("ACPI Managment: \n");
#endif
	printf("AC Line status: %s\n",
	    stat & BATT_AC_ONLINE ? "on-line" : "off-line");

	printf("Remaining battery life: ");
	if (valid_units) {
		printf("%d%%\n", total_cap / valid_units);
	} else {
		printf("unknown\n");
	}

	printf("Remaining battery time: ");
	if ((stat > 0) || (valid_rate == 0)) {
		printf("unknown\n");
	} else {
		printf("%d minutes\n", total_min);
	}

	printf("Number of batteries: %d\n", units);
	for (i = 0; i < units; i++) {
		printf("Battery %d:\n", i);

		printf("\tBattery status: %s\n", status[bat[i].bst.state]);
		if (bat[i].bst.state >= BATT_STAT_MAX)
			continue;

		printf("\tRemaining battery life: %d%%\n", bat[i].cap);

		printf("\tRemaining battery time: ");
		if ((stat > 0) || (bat[i].bst.rate == 0)) {
			printf("unknown\n");
		} else {
			printf("%d minutes\n", bat[i].min);
		}

		printf("\tBattery output: ");
		if (bat[i].bst.volt == 0xffffffff) {
			printf("unknown\n");
		} else {
			printf("%gV\n", bat[i].bst.volt / 1000.0);
		}

		/* Additional informations */
		if (verbose) {
			printf("\tBattery model : %s\n", bat[i].bif.model);
			printf("\tBattery serial: %s\n", bat[i].bif.serial);
			printf("\tBattery type  : %s\n", bat[i].bif.type);
			printf("\tBattery tech  : %s\n", bat[i].bif.btech > 0 ?
			    "Secondary (rechargeable)" :
			    "Primary (non-rechargeable)");
			printf("\tBattery oeminfo: %s\n", bat[i].bif.oeminfo);
		}

		/* Debugging informations */
		if (debug > 0) {
			char *u = power_units[bat[i].bif.unit & 0x01];

			printf("\t_BIF info\n");
			printf("\t  unit : %d (%s)\n", bat[i].bif.unit, u);
			printf("\t  dcap : %5d%s\n", bat[i].bif.dcap, u);
			printf("\t  lfcap: %5d%s\n", bat[i].bif.lfcap, u);
			printf("\t  wcap : %5d%s\n", bat[i].bif.wcap, u);
			printf("\t  lcap : %5d%s\n", bat[i].bif.lcap, u);
			printf("\t  gra1 : %5d%s\n", bat[i].bif.gra1, u);
			printf("\t  gra2 : %5d%s\n", bat[i].bif.gra2, u);
			printf("\t  dvol : %5dmV\n", bat[i].bif.dvol);
#ifdef DEBUG
			printf("\t  btech: %d\n", bat[i].bif.btech);
			printf("\t  model  : %s\n", bat[i].bif.model);
			printf("\t  serial : %s\n", bat[i].bif.serial);
			printf("\t  type   : %s\n", bat[i].bif.type);
			printf("\t  oeminfo: %s\n", bat[i].bif.oeminfo);
#endif
			printf("\t_BST info\n");
			printf("\t  state: %d\n", bat[i].bst.state);
			printf("\t  rate : %5d%s\n", bat[i].bst.rate, u);
			printf("\t  cap  : %5d%s\n", bat[i].bst.cap, u);
			printf("\t  volt : %5dmV\n", bat[i].bst.volt);
		}
	}
	exit(0);
}
